
/**
 * An array-based implementation of a min-priority queue.
 * Elements are assumed to be Comparable.
 *
 * @author Rui Meireles
 * @version 1.0
 */
public class MyArrayPriorityQueue<E extends Comparable<E>> implements MyPriorityQueue<E>{

    /* the array that holds the queue's elements */
    private E[] earray;
    /* Keeps track of the queue's size */
    private int size = 0;

    /* Default capacity */
    private static final int DEF_CAP = 10;
    
    /**
     * Iterator implementation for our array-based priority queue.
     * Written as an inner class so we can easily access 
     * the queue's private fields.
     */ 
    private class MyArrayPqueueItr implements MyIterator<E>{
        private int nextIdx = MyArrayPriorityQueue.this.size - 1; // head is first

        MyArrayPqueueItr(){ }

        /**
         * Indicates whether there are any unvisited elements in the
         * associated collection.
         * 
         * @return true if there are unvisited elements, false otherwise.
         */
        public boolean hasNext(){ return nextIdx >= 0; }

        /**
         * Returns the next element to be visited in the associated collection.
         * 
         * @return a previously unvisited element from the associated 
         * collection, or null if there are none. 
         */
        public E next(){
            E retval = null;
            if (this.hasNext()) retval = MyArrayPriorityQueue.this.earray[this.nextIdx--];
            return retval;
        }
    }
    
    /**
     * Constructor used to create an empty queue using 
     * an array with the specified capacity.
     * 
     * @param capacity initial array capacity.
     * @return an empty queue.
     */
    public MyArrayPriorityQueue(int capacity){
        
        this.earray = (E[]) new Comparable[capacity];
        /*
         * What is this single line of code doing?
         * 1. We are extending the Comparable interface in the class definition
         * 2. We are allocating memory for an array (the memory reference is in earray!)
         * 3. a) from right to left, the array has capacity elements
         * 3. b) the array will hold elements of type Comparable
         * 3. c) we are typecasting the array to the specific type, E, that we need.
         * 4. Because we are using generics, the typecast to E[] is correct.
         */
        
    }

    /**
     * Constructor used to create an empty queue using 
     * an array with the default capacity of 10.
     * 
     * @return an empty queue.
     */
    public MyArrayPriorityQueue(){ this(DEF_CAP); }

    // methods from the MyWorklist interface
    /**
     * Adds an element to the worklist.
     *
     * @param elem the element to be added.
     * @return true if the collection changed as a result of this addition; 
     * false otherwise.
     */
    public boolean add(E elem){

        if (this.size == this.earray.length){ // create space if needed
            E[] newEarray = (E[]) new Comparable[this.earray.length*2]; // new, larger, array
            for (int i=0; i < this.earray.length; i++) // copy everything over
                newEarray[i] = this.earray[i];
            this.earray = newEarray;
        }
        
        // insert new element in a way that maintains ordering invariant
        int elemIdx = this.size; // for starters, assume elem will be the new head        
        while (elemIdx > 0 && elem.compareTo(this.earray[elemIdx-1]) >= 0) // shift elements right as needed
            this.earray[elemIdx--] = this.earray[elemIdx]; // shift right by 1
        this.earray[elemIdx] = elem; // store the element where it belongs
        
        this.size++; // register one more element
        return true; // we never fail to add
    }

    /**
     * Removes the next item to be removed from the worklist, 
     * according to its removal rule. Since this is a min-priority queue,
     * the smallest element should be removed.
     * 
     * @return the item removed, or null if queue is empty.
     */
    public E remove(){
        E retval = null; // default value 
        
        if (this.size > 0){ // do we have something to remove?
            retval = this.earray[--this.size]; // smallest element is always at head of queue
            this.earray[this.size] = null; // null it out to help garbage collector do its job
        }
        
        return retval;
    }
    
    /**
     * Retrieves the next item to be removed from the worklist, 
     * according to its removal rule, without actually removing it.
     * Since this is a min-priority queue, the smallest element should 
     * be returned.
     * 
     * @return the next item to be removed, or null if queue is empty.
     */
    public E peek(){
        return this.size > 0 ? this.earray[this.size-1] : null; // smallest element is always at head of queue
    }    
    
    // methods from the MyCollection interface
       
    /**
     * Returns the number of elements currently in the collection.
     *
     * @return the number of elements in the collection.
     */
    public int size(){
        return this.size;
    }
    
    /**
     * Removes an element from the collection.
     *
     * @param elem the element to be removed from the collection.
     * @return true if the collection changed as a result of this addition; 
     * false otherwise.
     */
    public boolean remove(E elem){
        int foundIdx = -1;
        for (int i=0; i < this.size; i++){ // look for the item
            if (this.earray[i].equals(elem)){
                foundIdx = i;
                break;
            }
        }
        
        boolean found = foundIdx != -1;
        if (found){ // if elem found, left shift elements to the right of it to remove it while preserving order
            this.size--; // one fewer element
            for (int i=foundIdx; i < this.size; i++)
                this.earray[i] = this.earray[i+1];
           this.earray[this.size] = null; // help the garbage collector out
        }

        return found;
    }

    /**
     * Checks whether the element passed as an argument is present in the collection.
     *
     * @param elem the element whose presence is to be tested.
     * @return true if the element is present in the collection; false otherwise.
     */
    public boolean contains(E elem){
        // traverse queue looking for elem
        for (int i=0; i < this.size; i++)
            if (this.earray[i].equals(elem)) return true; // found the element of interest        

        return false; // if we got this far, it means nothing was found
    }
    
    /**
     * Returns an iterator that lets us traverse the collection.
     *
     * @return an iterator object associated with the collection.
     */
    public MyIterator<E> iterator(){
        return new MyArrayPqueueItr();
    }
}
